#include "linserver.h"
#include "message.h"
#include <signal.h>
#include <stdlib.h>
#include <setjmp.h>

QSocket server[MAX_SERVERS];
int totalServers;

void serviceClient(int index,int handle) {
	if (!config.isAllowed(server[index].sourcePort,server[index].targetHost)) {
		msgSend("Access denied for %s:%d",server[index].targetHost,
			server[index].targetPort);
		return;
	}

	if (!config.getServiceByConnection(server[index].sourcePort,server[index].targetHost)) {
		msgSend("No service specified for %s:%d",server[index].targetHost,
			server[index].targetPort);
		return;
	}

	msgSend("Service: [%s] for %s:%d",config.currentService->service,
		server[index].targetHost,server[index].targetPort);

	setuid(config.currentService->uid);	
	if (!strcmp(config.currentService->service,"telnet"))
		telnetService(handle);
	else if (!strcmp(config.currentService->service,"forward"))
		forwardService(false,handle,config.currentService->fields[0],
			atoi(config.currentService->fields[1]),(char*)NULL,0);
	else if (!strcmp(config.currentService->service,"socks5"))
		socksServer(handle);
	else if (!strcmp(config.currentService->service,"pop"))
		popService(false,handle,(char*)NULL,0);
	else if (!strcmp(config.currentService->service,"spop"))
		popService(true,handle,config.currentService->fields[2],
			atoi(config.currentService->fields[3]));
	else if (!strcmp(config.currentService->service,"stelnet"))
		socksClientService(handle,config.currentService->fields[2],
			atoi(config.currentService->fields[3]));
	else if (!strcmp(config.currentService->service,"program"))
		programService(handle,config.currentService->fields);
	else if (!strcmp(config.currentService->service,"sforward"))
		forwardService(true,handle,config.currentService->fields[0],
			atoi(config.currentService->fields[1]),
			config.currentService->fields[2],
			atoi(config.currentService->fields[3]));
	else if (!strcmp(config.currentService->service,"admin"))
			linAdmin(handle);

	msgSend("Connection Terminated: %s:%d",server[index].targetHost,server[index].targetPort);
}

//----------------------------------------------------------------------------------------------
// Server Signal Handler
fd_set fd,ready;
int maxfd=0;
sigjmp_buf reset_jump;

void reset(int) {
	signal(SIGHUP,reset);
	msgSend("Signal received. Reading configuration files again.");

	config.dispose();
	if (!config.read(CONFIG_FILE)) {
		msgSend("LinGate Server Terminating...");
		exit(1);
	}

	// Mark Services which were previously running but should now be removed
	for(int i=0;i<totalServers;i++)
		if (!config.get(server[i].sourcePort)) {
			msgSend("Closing service on port: %d",server[i].sourcePort);
			close(server[i].handle);
			server[i].close(); // Close Server
		}
			
	// Now remove entries which have been closed
	bool changes=true;
	while(changes) {
		changes=false;
		for(int i=0;i<totalServers;i++)
			if (server[i].handle==-1) {
				for(int j=i;j<totalServers-1;j++)
					server[j]=server[j+1];
				changes=true; totalServers--;
				break;
			}
	}

	// Now start services which have been added
	if (!config.rollback()) {
		msgSend("No services specified in Configuration file. Terminating");
		exit(1);
	}

	bool ok;
	do {
		ok=true;
		for(int i=0;i<totalServers;i++) 
			if (server[i].sourcePort==config.current->port) { 
				ok=false; break;
			}
		if (ok) {
			if (!server[totalServers].create(config.current->port)) 
				msgSend("Unable to bind to protocol port: %d",
					config.current->port);
			else {
				msgSend("Service started at port: %d",
					config.current->port);
				totalServers++;
			}
		}
	} while(config.next());

	FD_ZERO(&fd); maxfd=0;
	for(int i=0;i<totalServers;i++) {
		FD_SET(server[i].handle,&fd);
		if (server[i].handle>maxfd) maxfd=server[i].handle;
	}
	signal(SIGHUP,reset);
	siglongjmp(reset_jump,1);
}

main() {
	signal(SIGCHLD,SIG_IGN);
	if (fork()) return;
	msgSend("LinGate Server started");
	if (!config.read(CONFIG_FILE)) {
		msgSend("LinGate Server Terminating...");
		return 1;
	}
	totalServers=0;
	if (!config.rollback()) {
		msgSend("No services specified in Configuration file. Terminating");
		return 1;
	}

	FD_ZERO(&fd);
	do {
		if (!server[totalServers].create(config.current->port)) {
			msgSend("Unable to bind to protocol port: %d",
				config.current->port);
		}
		else {
			msgSend("Service started at port: %d",config.current->port);
			FD_SET(server[totalServers].handle,&fd);
			if (maxfd<server[totalServers].handle)
				maxfd=server[totalServers].handle;
			totalServers++;
		}
	} while(config.next());

	int newconnection;
	sigsetjmp(reset_jump,1);
	signal(SIGCHLD,SIG_IGN);
	signal(SIGHUP,reset);

	while(1) {
		bcopy(&fd,&ready,sizeof fd);
		select(maxfd+1,&ready,NULL,NULL,NULL);

		for(int i=0;i<totalServers;i++)
			if (FD_ISSET(server[i].handle,&ready)) {
				newconnection=server[i].accept();
				if (newconnection==-1) continue;
				msgSend("New connection from %s:%d",
					server[i].targetHost,server[i].targetPort);
				if (fork()==0) {
					serviceClient(i,newconnection);
					exit(0);
				}
				else close(newconnection);
			}
	}
}

