#include "ser.h"
#include "token.h"
#include "message.h"
#include <stdlib.h>

QConfigFile config;

bool QConfigFile::readGroup(char *filename,QGroup **top) {

	if (filename[0]!='.') {
		*top=new QGroup;
		(*top)->next=(QGroup*)NULL;
		strcpy((*top)->ip,filename);
		return true;
	}

	msgSend("Reading Group File: %s",filename);
	FILE *file=fopen(filename,"r");
	if (!file) {
		msgSend("Error Reading: %s",filename);
		return false;
	}
	*top=(QGroup*)NULL;
	QGroup *node;
	char buffer[256];

	while(!feof(file)) {
		buffer[0]=0;
		fgets(buffer,256,file);
		if (buffer[0]=='#') continue;
		QToken::normalize(buffer);
		node=new QGroup;
		strcpy(node->ip,buffer);
		node->next=*top;
		*top=node;
	}
	fclose(file);
	msgSend("Successful: %s",filename);
	return true;
}

// Return Value
//	true - Entry added successfully
// 	false - Entry has errors, so discarded

bool QConfigFile::parseAccessEntry(QAccess **top,char *buffer,QSocket *client) {
	QToken::normalize(buffer);
	if (buffer[0]=='#') return false;

	QToken token;
	QAccess *node;
	char buffer2[256];

	token.set(buffer,":");
	if (token.total()!=2) return false;

	node=new QAccess;
	token.get(buffer2);
	if (!strcmp(buffer2,"allow")) node->access=LIN_ALLOW;
	else if (!strcmp(buffer2,"deny")) node->access=LIN_DENY;
	else { delete node; return false; }

	token.get(buffer2);
	strcpy(node->ipFile,buffer2);
	if (client) {
		if (!readGroupServer(*client,&node->group,buffer2)) {
			delete node; return false;
		}
	}
	else {
		if (!readGroup(buffer2,&node->group)) {
			delete node; return false;
		}
	}
	node->next=*top;
	*top=node;
	return true;
}

bool QConfigFile::readAccess(char *filename,QAccess **top) {
	msgSend("Reading Access File: %s",filename);
	FILE *file=fopen(filename,"r");
	if (!file) {
		msgSend("Error Reading: %s",filename);
		return false;
	}
	*top=(QAccess*)NULL;
	char buffer[300];

	while(!feof(file)) {
		buffer[0]=0;
		fgets(buffer,300,file);
		parseAccessEntry(top,buffer);
	}
	fclose(file);
	msgSend("Successful: %s",filename);
	return true;
}

// Return Value
//	true - If entry added successfully
//	false - Entry had errors, so discarded

bool QConfigFile::parseServiceEntry(QService **top,char *buffer,QSocket *client) {
	QToken::normalize(buffer);
	if (buffer[0]=='#') return false;

	QService *node;
	QToken token;
	char buffer2[256];
	token.set(buffer,":");
	if (token.total()!=7) return false;

	node=new QService;
	token.get(buffer2);
	strcpy(node->ipFile,buffer2);
	if (client) {
		if (!readGroupServer(*client,&node->group,buffer2)) {
			delete node; return false;
		}
	}
	else {
		if (!readGroup(buffer2,&node->group)) {
			delete node; return false;
		}
	}

	token.get(buffer2);
	node->uid=atoi(buffer2);
	token.get(node->service);
	for(int i=0;i<4;i++)
		token.get(node->fields[i]);
	node->next=*top; *top=node;
	return true;
}

bool QConfigFile::readService(char *filename,QService **top) {
	msgSend("Reading Service File: %s",filename);

	FILE *file=fopen(filename,"r");
	if (!file) {
		msgSend("Error Reading: %s",filename);
		return false;
	}

	char buffer[2048];
	QService *node;
	*top=(QService*)NULL;
	QToken token;

	while(!feof(file)) {
		buffer[0]=0;
		fgets(buffer,2048,file);
		parseServiceEntry(top,buffer);
	}
	fclose(file);
	msgSend("Successful: %s",filename);
	return true;
}

bool QConfigFile::parseConfigEntry(char *buffer,QSocket *client) {
	QToken::normalize(buffer);
	if (buffer[0]=='#') return false;

	QToken token;
	char buffer2[1024];
	QConfig *node;

	token.set(buffer,":");
	if (token.total()!=6) return false;

	node=new QConfig;
	token.get(buffer2);
	node->port=atoi(buffer2);

	token.get(buffer2);
	token.strlwr(buffer2);
	if (!strcmp(buffer2,"tcp")) node->protocol=LIN_TCP;
	else if (!strcmp(buffer2,"udp")) node->protocol=LIN_UDP;
	else { delete node; return false; }

	token.get(buffer2);
	token.strlwr(buffer2);
	if (!strcmp(buffer2,"wait")) node->serverType=LIN_WAIT;
	else if (!strcmp(buffer2,"nowait")) node->serverType=LIN_NOWAIT;
	else { delete node; return false; }

	token.get(buffer2);
	node->backlog=atoi(buffer2);

	token.get(node->accessFile);
	token.get(node->serviceFile);

	if (node->accessFile[0]) {
		if (!client) readAccess(node->accessFile,&node->access);
		else readAccessServer(*client,&node->access,node->port,
			node->protocol==LIN_UDP ? "udp" : "tcp");
	}

	if (!client) {
		if (!readService(node->serviceFile,&node->service)) {
			delete node; return false;	
		}
	}
	else {
		if (!readServiceServer(*client,&node->service,node->port,
			node->protocol==LIN_UDP ? "udp" : "tcp"))
		{
			delete node; return false;
		}
	}
	node->next=config;
	config=node;
	return true;
}

bool QConfigFile::read(char *filename) {
	msgSend("Reading Configuration File: %s",filename);
	FILE *file=fopen(filename,"r");
	if (!file) {
		msgSend("Error Reading: %s",filename);
		return false;
	}

	char buffer[2048];
	while(!feof(file)) {
		buffer[0]=0;
		fgets(buffer,2048,file);
		parseConfigEntry(buffer);
	}
	fclose(file);
	msgSend("Successful: %s",filename);
	return true;
}		



void QConfigFile::display() {
	QConfig *c;
	QAccess *a;
	QService *s;
	QGroup *g;

	c=config;
	while(c) {
		printf("Source Port: %d\nProtocol: %d\nServer Type: %d\nBacklog: %d"
			"\nAccess File: %s\nService File: %s\n", c->port,c->protocol,
			c->serverType,c->backlog,c->accessFile,c->serviceFile);
		a=c->access;
		while(a) {
			printf("\tAcces: %d\n\tIP File: %s\n",a->access,a->ipFile);
			g=a->group;
			while(g) {
				printf("\t\tIP: %s\n",g->ip); g=g->next;
			}
			a=a->next;
		}
		s=c->service;
		while(s) {
			printf("\tIP File: %s\n\tUID: %d\n\tService: %s\n",s->ipFile,
				s->uid,s->service);
			g=s->group;
			while(g) {
				printf("\t\tIP: %s\n",g->ip); g=g->next;
			}
			s=s->next;
		}
		c=c->next;
	}
	printf("\n");
}

bool QConfigFile::rollback() {
	if (!config) return false;
	current=config;
	return true;
}

bool QConfigFile::next() {
	if (!current || !current->next) return false;
	current=current->next;
	return true;
}

bool QConfigFile::rollbackAccess() {
	if (!current->access) return false;
	currentAccess=current->access;
	return true;
}

bool QConfigFile::nextAccess() {
	if (!currentAccess || !currentAccess->next) return false;
	currentAccess=currentAccess->next;
	return true;
}

bool QConfigFile::rollbackService() {
	if (!current->service) return false;
	currentService=current->service;
	return true;
}

bool QConfigFile::nextService() {
	if (!currentService || !currentService->next) return false;
	currentService=currentService->next;
	return true;
}

bool QConfigFile::isMatch(char *pattern,char *ip) {
	while(*pattern && *ip) {
		if (*pattern==*ip) { ip++; pattern++; }
		else if (*pattern=='*') {
			pattern++;
			while(*ip && *ip!='.') ip++;
		}
		else return false;
	}
	if (!*pattern && !*ip) return true;
	return false;
}

bool QConfigFile::get(int port,int protocol) {
	current=config;
	while(current) {
		if (current->port==port && current->protocol==protocol)
break;
		current=current->next;
	}
	return current ? true : false;
}

bool QConfigFile::isAllowed(int port,char *host) {
	if (!get(port)) return false;

	if (!current->access) return true;

	QAccess *a=current->access;
	QGroup *g;
	while(a) {
		g=a->group;
		while(g) {
			if (isMatch(g->ip,host)) 
				return a->access==LIN_ALLOW?true:false;
			g=g->next;
		}
		a=a->next;
	}
	return false;
}

bool QConfigFile::getServiceByConnection(int port,char *host) {
	if (!get(port)) return false;
	
	QGroup *g;
	currentService=current->service;
	if (!currentService) return false;

	while(currentService) {
		g=currentService->group;
		while(g) {
			if (isMatch(g->ip,host)) return true;
			g=g->next;
		}
		currentService=currentService->next;
	}
	return false;
}

void QConfigFile::disposeGroup(QGroup **top) {
	QGroup *p,*q;
	p=q=*top;
	while(p) {
		p=q->next;
		delete q;
		q=p;
	}
	*top=(QGroup*)NULL;
}

void QConfigFile::disposeAccess(QAccess **top) {
	QAccess *p,*q;
	p=q=*top;
	while(p) {
		if (p->group) disposeGroup(&p->group);
		p=q->next;
		delete q;
		q=p;
	}
	*top=(QAccess*)NULL;
}

void QConfigFile::disposeService(QService **top) {
	QService *p,*q;
	p=q=*top;
	while(p) {
		if (p->group) disposeGroup(&p->group);
		p=q->next;
		delete q;
		q=p;
	}
	*top=(QService*)NULL;
}

void QConfigFile::dispose() {
	QConfig *p,*q;
	p=q=config;
	while(p) {
		if (p->access) disposeAccess(&p->access);
		if (p->service) disposeService(&p->service);
		p=q->next;
		delete q;
		q=p;
	}
	current=config=(QConfig*)NULL;
	currentService=(QService*)NULL;
	currentAccess=(QAccess*)NULL;
}

bool QConfigFile::readGroupServer(QSocket &client,QGroup **top,char *name) {
	char buffer[2048];

	if (name[0]!='.') {
		*top=new QGroup;
		(*top)->next=(QGroup*)NULL;
		strcpy((*top)->ip,name);
		return true;
	}

	sprintf(buffer,"listgroup %s\r\n",name);
	client.write(buffer);
	QGroup *node;

	while(1) {
		client.readline(buffer,2048);
		if (buffer[0]=='*') break;
		else if (buffer[0]!='+') return false;
		QToken::normalize(buffer);

		node=new QGroup;
		node->next=*top;
		strcpy(node->ip,buffer);
		*top=node;
	}
	return true;
}
				
bool QConfigFile::readAccessServer(QSocket &client,QAccess **top,int port,char *protocol) {
	char buffer[2048];
	
	sprintf(buffer,"listaccess %d %s\r\n",port,protocol);
	client.write(buffer);
	*top=(QAccess*)NULL;

	while(1) {
		client.readline(buffer,2048);
		printf("readAccessServer: %s\n",buffer);
		if (buffer[0]=='*') break;
		if (buffer[0]!='+') return false;
		parseAccessEntry(top,&buffer[1],&client);
	}
	return true;
}

bool QConfigFile::readServiceServer(QSocket &client,QService **top,int port,char *protocol) {
	char buffer[2048];
	sprintf(buffer,"listservice %d %s\r\n",port,protocol);
	client.write(buffer);

	*top=(QService*)NULL;
	while(1) {
		client.readline(buffer,2048);
		printf("readServiceServer: %s\n",buffer);
		if (buffer[0]=='*') break;
		if (buffer[0]!='+') return false;
		parseServiceEntry(top,&buffer[1],&client);
	}
	return true;
}

bool QConfigFile::readServer(QSocket &client) {
	char buffer[2048];

	config=(QConfig*)NULL;
	try {
		client.write("rollback\r\n");
		client.readline(buffer,1024);
		if (buffer[0]=='-') throw QSocketException();
		while(1) {
			client.write("next\r\n");
			client.readline(buffer,1024);
			printf("readServer: %s\n",buffer);
			if (buffer[0]=='-') break;
			if (buffer[0]!='+') {
				client.close();
				return false;
			}
			parseConfigEntry(&buffer[1],&client);
		}
	}
	catch(...) {
		client.close();
		return false;
	}
	return true;
}

